/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 * 
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License. You can obtain
 * a copy of the License at https://glassfish.dev.java.net/public/CDDL+GPL.html
 * or glassfish/bootstrap/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 * 
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at glassfish/bootstrap/legal/LICENSE.txt.
 * Sun designates this particular file as subject to the "Classpath" exception
 * as provided by Sun in the GPL Version 2 section of the License file that
 * accompanied this code.  If applicable, add the following below the License
 * Header, with the fields enclosed by brackets [] replaced by your own
 * identifying information: "Portions Copyrighted [year]
 * [name of copyright owner]"
 * 
 * Contributor(s):
 * 
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */
package com.sun.grizzly.jruby.rack;

import com.sun.grizzly.jruby.RailsAdapter;
import com.sun.grizzly.tcp.http11.GrizzlyRequest;
import com.sun.grizzly.tcp.http11.GrizzlyResponse;
import org.jruby.Ruby;
import org.jruby.exceptions.RaiseException;
import org.jruby.javasupport.JavaEmbedUtils;
import org.jruby.runtime.builtin.IRubyObject;

import java.io.IOException;
import java.util.Collections;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * @author Vivek Pandey
 */
public abstract class DefaultRackApplicationFactory implements RackApplicationFactory{
    private RackApplication errorApplication;
    protected final Logger logger;
    final String appRoot;
    private final RailsAdapter adapter;


    public DefaultRackApplicationFactory(RailsAdapter adapter) {
        this.logger = adapter.getLogger();
        this.appRoot = adapter.getAppRoot();
        this.adapter = adapter;
    }

    public RackApplication newApplication(final Ruby runtime) throws RackInitializationException {
        return createApplication(new ApplicationObjectFactory() {
            public IRubyObject create() {
                return createApplicationObject(runtime);
            }

            public Ruby runtime() {
                return runtime;
            }
        });
    }

    public void finishedWithApplication(RackApplication app) {
        app.destroy();
    }

    public RackApplication getErrorApplication(Ruby runtime) {
        if(errorApplication == null){
            errorApplication = createErrorApp(runtime);
        }
        return errorApplication;
    }

    public void destroy() {
        if(errorApplication != null)
            errorApplication.destroy();
        errorApplication = null;
    }

    public abstract IRubyObject createApplicationObject(Ruby runtime);

    public IRubyObject createErrorApplicationObject(Ruby runtime) {
        return runtime.evalScriptlet(
                "Rack::Handler::Grizzly.new(Rack::Builder.new {( "
                + "run JRuby::Rack::ErrorsApp.new" + "\n )}.to_app)");
    }

    private RackApplication createErrorApp(final Ruby runtime){
        try {
            RackApplication app =
                    createApplication(new ApplicationObjectFactory() {
                public IRubyObject create() {
                    return createErrorApplicationObject(runtime);
                }

                        public Ruby runtime() {
                            return runtime;
                        }
                    });

            return app;
        } catch (final Exception e) {
            logger.log(Level.WARNING,
                "Warning: error application could not be initialized.");
            return new DefaultRackApplication(createErrorApplicationObject(runtime), adapter) {
                public void init() throws RackInitializationException { }
                public RackResponse call(GrizzlyRequest env) {
                    return new RackResponse() {
                        public int getStatus() { return 500; }
                        public Map getHeaders() { return Collections.EMPTY_MAP; }
                        public String getBody() { return ""; }
                        public void respond(GrizzlyResponse response) {
                            try {
                                response.sendError(500,
                                    "Application initialization failed: "
                                    + e.getMessage());
                            } catch (IOException ex) {
                                logger.severe("Unable to send default error page!");
                            }
                        }
                    };
                }
                public void destroy() {
                    JavaEmbedUtils.terminate(runtime);
                }
            };
        }
    }

    public RackApplication newErrorApplication(Ruby runtime) {
        if(errorApplication == null){
            errorApplication = createErrorApp(runtime);
        }
        return errorApplication;
    }

    private interface ApplicationObjectFactory {
        IRubyObject create();
        Ruby runtime();
    }

    private RackApplication createApplication(final ApplicationObjectFactory appfact)
            throws RackInitializationException {
        try {
            return new DefaultRackApplication(appfact.create(), adapter) {
                @Override
                public void destroy() {
                    JavaEmbedUtils.terminate(appfact.runtime());
                }
            };
        } catch (RaiseException re) {
            throw new RackInitializationException(re);
        }
    }
}
